# -*- coding: utf-8 -*-

# Copyright (c) 2018 Minoru Osuka
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 		http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from cockatrice.analysis import JanomeTokenizer


class TestJanomeTokenizer(unittest.TestCase):
    def test_janome_tokenizer(self):
        value = u'すもももももももものうち'

        tokenizer = JanomeTokenizer()

        expected = ['すもも', 'も', 'もも', 'も', 'もも', 'の', 'うち']
        actual = [t.text for t in tokenizer(value)]

        self.assertEqual(expected, actual)
